from unittest.mock import patch
import pytest
import requests

from piffle import image


api_endpoint = "http://imgserver.co"
image_id = "img1"

# test iiif image urls
VALID_URLS = {
    "info": "%s/%s/info.json" % (api_endpoint, image_id),
    # longer api endpoint path
    "info-loris": "%s/loris/%s/info.json" % (api_endpoint, image_id),
    "simple": "%s/%s/full/full/0/default.jpg" % (api_endpoint, image_id),
    "complex": "%s/%s/2560,2560,256,256/256,/!90/default.jpg"
    % (api_endpoint, image_id),
    "exact": "%s/%s/full/!256,256/0/default.jpg" % (api_endpoint, image_id),
}

INVALID_URLS = {
    "info": "http://img1/info.json",
    "simple": "http://imgserver.co/img1/foobar/default.jpg",
    "complex": "http://imgserver.co/img1/2560,2560,256,/256,/!90/default.jpg",
    "bad_size": "%s/%s/full/a,/0/default.jpg" % (api_endpoint, image_id),
    "bad_region": "%s/%s/200,200/full/0/default.jpg" % (api_endpoint, image_id),
}

sample_image_info = {
    "@context": "http://image.io/api/image/2/context.json",
    "@id": VALID_URLS["simple"],
    "height": 3039,
    "width": 2113,
}


def get_test_imgclient():
    return image.IIIFImageClient(api_endpoint=api_endpoint, image_id=image_id)


class TestIIIFImageClient:
    def test_defaults(self):
        img = get_test_imgclient()
        # default image url
        assert "%s/%s/full/full/0/default.jpg" % (api_endpoint, image_id) == str(img)
        # info url
        assert "%s/%s/info.json" % (api_endpoint, image_id) == str(img.info())

    def test_outputs(self):
        img = get_test_imgclient()
        # str and unicode should be equivalent
        assert str(img.info()) == str(str(img.info()))
        # repr should have class and image id
        assert "IIIFImageClient" in repr(img)
        assert img.get_image_id() in repr(img)

    def test_init_opts(self):
        test_opts = {
            "region": "2560,2560,256,256",
            "size": "256,",
            "rotation": "90",
            "quality": "color",
            "fmt": "png",
        }
        img = image.IIIFImageClient(
            api_endpoint=api_endpoint, image_id=image_id, **test_opts
        )
        assert img.api_endpoint == api_endpoint
        assert img.image_id == image_id
        expected_img_opts = test_opts.copy()
        del expected_img_opts["region"]
        del expected_img_opts["size"]
        del expected_img_opts["rotation"]
        assert img.image_options == expected_img_opts
        assert str(img.region) == test_opts["region"]
        assert str(img.size) == test_opts["size"]
        assert str(img.rotation) == test_opts["rotation"]

        # TODO: should parse/verify options on init
        # with pytest.raises(image.IIIFImageClientException):
        #     img = image.IIIFImageClient(api_endpoint=api_endpoint, fmt='bogus')

    def test_size(self):
        img = get_test_imgclient()
        width, height, percent = 100, 150, 50
        # width only
        assert "%s/%s/full/%s,/0/default.jpg" % (api_endpoint, image_id, width) == str(
            img.size(width=width)
        )

        # height only
        assert "%s/%s/full/,%s/0/default.jpg" % (api_endpoint, image_id, height) == str(
            img.size(height=height)
        )
        # width and height
        assert "%s/%s/full/%s,%s/0/default.jpg" % (
            api_endpoint,
            image_id,
            width,
            height,
        ) == str(img.size(width=width, height=height))
        # exact width and height
        assert "%s/%s/full/!%s,%s/0/default.jpg" % (
            api_endpoint,
            image_id,
            width,
            height,
        ) == str(img.size(width=width, height=height, exact=True))
        # percent
        assert "%s/%s/full/pct:%s/0/default.jpg" % (
            api_endpoint,
            image_id,
            percent,
        ) == str(img.size(percent=percent))

    def test_region(self):
        # region options passed through to region object and output
        img = get_test_imgclient()
        x, y, width, height = 5, 10, 100, 150
        assert "%s/%s/%s,%s,%s,%s/full/0/default.jpg" % (
            api_endpoint,
            image_id,
            x,
            y,
            width,
            height,
        ) == str(img.region(x=x, y=y, width=width, height=height))

    def test_rotation(self):
        # rotation options passed through to region object and output
        img = get_test_imgclient()
        assert "%s/%s/full/full/90/default.jpg" % (api_endpoint, image_id) == str(
            img.rotation(degrees=90)
        )
        with pytest.raises(image.IIIFImageClientException):
            img.rotation(foo="bar")

    def test_format(self):
        img = get_test_imgclient()
        png = img.format("png")
        jpg = img.format("jpg")
        gif = img.format("gif")
        assert str(png).endswith(".png")
        assert str(jpg).endswith(".jpg")
        assert str(gif).endswith(".gif")

        with pytest.raises(image.IIIFImageClientException):
            img.format("bogus")

    def test_combine_options(self):
        img = get_test_imgclient()
        width = 100
        fmt = "png"
        assert "%s/%s/full/%s,/0/default.%s" % (
            api_endpoint,
            image_id,
            width,
            fmt,
        ) == str(img.size(width=width).format(fmt))

        img = get_test_imgclient()
        x, y, width, height = 5, 10, 100, 150
        assert "%s/%s/%s,%s,%s,%s/full/0/default.%s" % (
            api_endpoint,
            image_id,
            x,
            y,
            width,
            height,
            fmt,
        ) == str(img.region(x=x, y=y, width=width, height=height).format(fmt))

        assert "%s/%s/%s,%s,%s,%s/%s,/0/default.%s" % (
            api_endpoint,
            image_id,
            x,
            y,
            width,
            height,
            width,
            fmt,
        ) == str(
            img.size(width=width)
            .region(x=x, y=y, width=width, height=height)
            .format(fmt)
        )
        rotation = 90
        assert "%s/%s/%s,%s,%s,%s/%s,/%s/default.%s" % (
            api_endpoint,
            image_id,
            x,
            y,
            width,
            height,
            width,
            rotation,
            fmt,
        ) == str(
            img.size(width=width)
            .region(x=x, y=y, width=width, height=height)
            .rotation(degrees=90)
            .format(fmt)
        )

        # original image object should be unchanged, and still show defaults
        assert "%s/%s/full/full/0/default.jpg" % (api_endpoint, image_id) == str(img)

    def test_init_from_url(self):
        # well-formed
        # - info url
        img = image.IIIFImageClient.init_from_url(VALID_URLS["info"])
        assert isinstance(img, image.IIIFImageClient)
        assert img.image_id == image_id
        # round trip back to original url
        assert str(img.info()) == VALID_URLS["info"]
        assert img.api_endpoint == api_endpoint
        # -info with more complex endpoint base url
        img = image.IIIFImageClient.init_from_url(VALID_URLS["info-loris"])
        assert img.api_endpoint == "%s/loris" % api_endpoint
        # - image
        img = image.IIIFImageClient.init_from_url(VALID_URLS["simple"])
        assert isinstance(img, image.IIIFImageClient)
        assert str(img) == VALID_URLS["simple"]
        img = image.IIIFImageClient.init_from_url(VALID_URLS["complex"])
        assert str(img) == VALID_URLS["complex"]
        assert isinstance(img, image.IIIFImageClient)
        img = image.IIIFImageClient.init_from_url(VALID_URLS["exact"])
        assert str(img) == VALID_URLS["exact"]
        assert isinstance(img, image.IIIFImageClient)
        assert img.size.options["exact"] is True

        # malformed
        with pytest.raises(image.ParseError):
            img = image.IIIFImageClient.init_from_url(INVALID_URLS["info"])
        with pytest.raises(image.ParseError):
            image.IIIFImageClient.init_from_url(INVALID_URLS["simple"])
        with pytest.raises(image.ParseError):
            image.IIIFImageClient.init_from_url(INVALID_URLS["complex"])
        with pytest.raises(image.ParseError):
            image.IIIFImageClient.init_from_url(INVALID_URLS["bad_size"])
        with pytest.raises(image.ParseError):
            image.IIIFImageClient.init_from_url(INVALID_URLS["bad_region"])
        with pytest.raises(image.ParseError):
            image.IIIFImageClient.init_from_url("http://info.json")

    def test_as_dicts(self):
        img = image.IIIFImageClient.init_from_url(VALID_URLS["complex"])
        assert img.as_dict() == {
            "region": {
                "full": False,
                "square": False,
                "height": 256,
                "percent": False,
                "width": 256,
                "x": 2560,
                "y": 2560,
            },
            "rotation": {"degrees": 90.0, "mirrored": True},
            "size": {
                "exact": False,
                "full": False,
                "max": False,
                "height": None,
                "percent": None,
                "width": 256,
            },
            "quality": "default",
            "format": "jpg",
        }

    @patch("piffle.image.requests")
    def test_image_info(self, mockrequests):
        # test image info logic by mocking requests
        mockrequests.codes.ok = requests.codes.ok
        mockresponse = mockrequests.get.return_value
        mockresponse.status_code = requests.codes.ok
        mockresponse.json.return_value = sample_image_info

        # valid response
        img = image.IIIFImageClient.init_from_url(VALID_URLS["simple"])
        assert img.image_info == sample_image_info
        mockrequests.get.assert_called_with(img.info())
        mockresponse.json.assert_called_with()

        # error response
        mockresponse.status_code = 400
        img = image.IIIFImageClient.init_from_url(VALID_URLS["simple"])
        img.image_info
        mockresponse.raise_for_status.assert_called_with()

    def test_image_width_height(self):
        img = image.IIIFImageClient.init_from_url(VALID_URLS["simple"])

        with patch.object(image.IIIFImageClient, "image_info", new=sample_image_info):
            assert img.image_width == sample_image_info["width"]
            assert img.image_height == sample_image_info["height"]

    def test_canonicalize(self):
        img = image.IIIFImageClient.init_from_url(VALID_URLS["simple"])
        square_img_info = sample_image_info.copy()
        square_img_info.update({"height": 100, "width": 100})
        with patch.object(image.IIIFImageClient, "image_info", new=square_img_info):
            # square region for square image = full
            img.region.parse("square")
            # percentage: convert to w,h (= 25,25)
            img.size.parse("pct:25")
            img.rotation.parse("90.0")
            assert str(img.canonicalize()) == "%s/%s/full/25,25/90/default.jpg" % (
                api_endpoint,
                image_id,
            )


class TestImageRegion:
    def test_defaults(self):
        region = image.ImageRegion()
        assert str(region) == "full"
        assert region.as_dict() == image.ImageRegion.region_defaults

    def test_init(self):
        # full
        region = image.ImageRegion(full=True)
        assert region.as_dict()["full"] is True
        # square
        region = image.ImageRegion(square=True)
        assert region.as_dict()["square"] is True
        assert region.as_dict()["full"] is False
        # region
        region = image.ImageRegion(x=5, y=7, width=100, height=103)
        assert region.as_dict()["full"] is False
        assert region.as_dict()["x"] == 5
        assert region.as_dict()["y"] == 7
        assert region.as_dict()["width"] == 100
        assert region.as_dict()["height"] == 103
        assert region.as_dict()["percent"] is False
        # percentage region
        region = image.ImageRegion(x=5, y=7, width=100, height=103, percent=True)
        assert region.as_dict()["full"] is False
        assert region.as_dict()["x"] == 5
        assert region.as_dict()["y"] == 7
        assert region.as_dict()["width"] == 100
        assert region.as_dict()["height"] == 103
        assert region.as_dict()["percent"] is True

        # errors
        with pytest.raises(image.IIIFImageClientException):
            # invalid parameter
            image.ImageRegion(bogus="foo")

            # incomplete options
            image.ImageRegion(x=1)
            image.ImageRegion(x=1, y=2)
            image.ImageRegion(x=1, y=2, w=20)
            image.ImageRegion(percent=True)
            image.ImageRegion().set_options(percent=True, x=1)

            # TODO: type checking? (not yet implemented)

        with pytest.raises(image.ParseError):
            image.ImageRegion().parse("1,2")

    def test_render(self):
        region = image.ImageRegion(full=True)
        assert str(region) == "full"
        region = image.ImageRegion(square=True)
        assert str(region) == "square"
        region = image.ImageRegion(x=5, y=5, width=100, height=100)
        assert str(region) == "5,5,100,100"
        region = image.ImageRegion(x=5, y=5, width=100, height=100, percent=True)
        assert str(region) == "pct:5,5,100,100"
        region = image.ImageRegion(
            x=5.1, y=3.14, width=100.76, height=100.89, percent=True
        )
        assert str(region) == "pct:5.1,3.14,100.76,100.89"

    def test_parse(self):
        region = image.ImageRegion()
        # full
        region_str = "full"
        region.parse(region_str)
        assert str(region) == region_str  # round trip
        assert region.as_dict()["full"] is True
        # square
        region_str = "square"
        region.parse(region_str)
        assert str(region) == region_str  # round trip
        assert region.as_dict()["full"] is False
        assert region.as_dict()["square"] is True
        # region
        x, y, w, h = [5, 7, 100, 200]
        region_str = "%d,%d,%d,%d" % (x, y, w, h)
        region.parse(region_str)
        assert str(region) == region_str  # round trip
        region_opts = region.as_dict()
        assert region_opts["full"] is False
        assert region_opts["square"] is False
        assert region_opts["x"] == x
        assert region_opts["y"] == y
        assert region_opts["width"] == w
        assert region_opts["height"] == h
        # percentage region
        region_str = "pct:%d,%d,%d,%d" % (x, y, w, h)
        region.parse(region_str)
        assert str(region) == region_str  # round trip
        region_opts = region.as_dict()
        assert region_opts["full"] is False
        assert region_opts["square"] is False
        assert region_opts["x"] == x
        assert region_opts["y"] == y
        assert region_opts["width"] == w
        assert region_opts["height"] == h
        assert region_opts["percent"] is True

        # invalid or incomplete region strings
        with pytest.raises(image.ParseError):
            region.parse("pct:1,3,")
        with pytest.raises(image.ParseError):
            region.parse("one,two,three,four")

    def test_canonicalize(self):
        # any canonicalization that requires image dimensions to calculate
        # should raise an error
        region = image.ImageRegion()
        region.parse("square")
        with pytest.raises(image.IIIFImageClientException):
            region.canonicalize()

        img = image.IIIFImageClient.init_from_url(VALID_URLS["simple"])
        # full to full - trivial canonicalization
        img.region.canonicalize()
        assert str(img.region) == "full"
        # x,y,w,h should be preserved as is
        dimensions = "0,0,200,250"
        img.region.parse(dimensions)
        img.region.canonicalize()
        # round trip, should be the same
        assert str(img.region) == dimensions

        # test with square image size
        square_img_info = sample_image_info.copy()
        square_img_info.update({"height": 100, "width": 100})
        with patch.object(image.IIIFImageClient, "image_info", new=square_img_info):
            # square requested, image is square = full
            img.region.parse("square")
            img.region.canonicalize()
            assert str(img.region) == "full"

            # percentages
            img.region.parse("pct:10,1,50,75")
            img.region.canonicalize()
            assert str(img.region) == "10,1,50,75"

            # percentages should be converted to integers
            img.region.parse("pct:10,1,50.5,75.3")
            assert str(img.region) == "pct:10,1,50.5,75.3"
            img.region.canonicalize()
            assert str(img.region) == "10,1,50,75"

        # test with square with non-square image size
        tall_img_info = sample_image_info.copy()
        tall_img_info.update({"width": 100, "height": 150})
        with patch.object(image.IIIFImageClient, "image_info", new=tall_img_info):
            # square requested, should convert to x,y,w,h
            img.region.parse("square")
            img.region.canonicalize()
            assert str(img.region) == "0,25,100,100"

        wide_img_info = sample_image_info.copy()
        wide_img_info.update({"width": 200, "height": 50})
        with patch.object(image.IIIFImageClient, "image_info", new=wide_img_info):
            # square requested, should convert to x,y,w,h
            img.region.parse("square")
            img.region.canonicalize()

            assert str(img.region) == "75,0,50,50"


class TestImageSize:
    def test_defaults(self):
        size = image.ImageSize()
        assert str(size) == "full"
        assert size.as_dict() == image.ImageSize.size_defaults

    def test_init(self):
        # full
        size = image.ImageSize(full=True)
        assert size.as_dict()["full"] is True
        # max
        size = image.ImageSize(max=True)
        assert size.as_dict()["max"] is True
        assert size.as_dict()["full"] is False
        # percentage
        size = image.ImageSize(percent=50)
        assert size.as_dict()["full"] is False
        assert size.as_dict()["percent"] == 50
        # width only
        size = image.ImageSize(width=100)
        assert size.as_dict()["width"] == 100
        # height only
        size = image.ImageSize(height=200)
        assert size.as_dict()["height"] == 200

        # errors
        with pytest.raises(image.IIIFImageClientException):
            # invalid parameter
            image.ImageSize(bogus="foo")

            # incomplete options ?
            # type checking? (not yet implemented)

    def test_render(self):
        size = image.ImageSize(full=True)
        assert str(size) == "full"
        size = image.ImageSize(max=True)
        assert str(size) == "max"
        size = image.ImageSize(percent=50)
        assert str(size) == "pct:50"
        size = image.ImageSize(width=100, height=105)
        assert str(size) == "100,105"
        size = image.ImageSize(width=100)
        assert str(size) == "100,"
        size = image.ImageSize(height=105)
        assert str(size) == ",105"

    def test_parse(self):
        size = image.ImageSize()
        # full
        size_str = "full"
        size.parse(size_str)
        assert str(size) == size_str  # round trip
        assert size.as_dict()["full"] is True
        # max
        size_str = "max"
        size.parse(size_str)
        assert str(size) == size_str  # round trip
        assert size.as_dict()["full"] is False
        assert size.as_dict()["max"] is True
        # width and height
        w, h = [100, 200]
        size_str = "%d,%d" % (w, h)
        size.parse(size_str)
        assert str(size) == size_str  # round trip
        size_opts = size.as_dict()
        assert size_opts["full"] is False
        assert size_opts["max"] is False
        assert size_opts["width"] == w
        assert size_opts["height"] == h
        # percentage size
        size_str = "pct:55"
        size.parse(size_str)
        assert str(size) == size_str  # round trip
        size_opts = size.as_dict()
        assert size_opts["full"] is False
        assert size_opts["percent"] == 55

        # invalid or incomplete size strings
        with pytest.raises(image.ParseError):
            size.parse("pct:")
        with pytest.raises(image.ParseError):
            size.parse("one,two")

    def test_canonicalize(self):
        # any canonicalization that requires image dimensions to calculate
        # should raise an error
        size = image.ImageSize()
        size.parse(",5")
        with pytest.raises(image.IIIFImageClientException):
            size.canonicalize()

        img = image.IIIFImageClient.init_from_url(VALID_URLS["simple"])
        # full to full - trivial canonicalization
        img.size.canonicalize()
        assert str(img.size) == "full"

        # test sizes with square image size
        square_img_info = sample_image_info.copy()
        square_img_info.update({"height": 100, "width": 100})
        with patch.object(image.IIIFImageClient, "image_info", new=square_img_info):
            # requested as ,h - convert to w,
            img.size.parse(",50")
            img.size.canonicalize()
            assert str(img.size) == "50,"

            # percentage: convert to w,h
            img.size.parse("pct:25")
            img.size.canonicalize()
            assert str(img.size) == "25,25"

            # exact
            img.size.parse("!50,50")
            img.size.canonicalize()
            assert str(img.size) == "50,50"

        # test sizes with rectangular image size
        rect_img_info = sample_image_info.copy()
        rect_img_info.update({"width": 50, "height": 100})
        with patch.object(image.IIIFImageClient, "image_info", new=rect_img_info):
            img.size.parse("!50,50")
            img.size.canonicalize()
            assert str(img.size) == "25,50"


class TestImageRotation:
    def test_defaults(self):
        rotation = image.ImageRotation()
        assert str(rotation) == "0"
        assert rotation.as_dict() == image.ImageRotation.rotation_defaults

    def test_init(self):
        # degrees
        rotation = image.ImageRotation(degrees=90)
        assert rotation.as_dict()["degrees"] == 90
        assert rotation.as_dict()["mirrored"] is False
        rotation = image.ImageRotation(degrees=95, mirrored=True)
        assert rotation.as_dict()["degrees"] == 95
        assert rotation.as_dict()["mirrored"] is True

    def test_render(self):
        rotation = image.ImageRotation()
        assert str(rotation) == "0"
        rotation = image.ImageRotation(degrees=90)
        assert str(rotation) == "90"
        rotation = image.ImageRotation(degrees=95, mirrored=True)
        assert str(rotation) == "!95"

        # canonicalization
        # - trim any trailing zeros in a decimal value
        assert str(image.ImageRotation(degrees=93.0)) == "93"
        # - leading zero if less than 1
        assert str(image.ImageRotation(degrees=0.05)) == "0.05"
        # - ! if mirrored, followed by integer if possible
        rotation = image.ImageRotation(degrees=95.00, mirrored=True)
        assert str(rotation) == "!95"
        # explicitly test canonicalize method, even though it does nothing
        rotation.canonicalize()
        assert str(rotation) == "!95"

    def test_parse(self):
        rotation = image.ImageRotation()
        rotation_str = "180"
        rotation.parse(rotation_str)
        assert str(rotation) == rotation_str  # round trip
        rotation_str = "!90"
        rotation.parse(rotation_str)
        assert str(rotation) == rotation_str  # round trip
        assert rotation.as_dict()["mirrored"] is True



import json
import os.path
from unittest.mock import patch

import pytest
import requests

from piffle.presentation import IIIFPresentation, IIIFException


FIXTURE_DIR = os.path.join(os.path.dirname(__file__), "fixtures")


class TestIIIFPresentation:
    test_manifest = os.path.join(FIXTURE_DIR, "chto-manifest.json")

    def test_from_file(self):
        pres = IIIFPresentation.from_file(self.test_manifest)
        assert isinstance(pres, IIIFPresentation)
        assert pres.type == "sc:Manifest"

    def test_from_url(self):
        manifest_url = "http://ma.ni/fe.st"
        with open(self.test_manifest) as manifest:
            data = json.loads(manifest.read())
        with patch("piffle.presentation.requests") as mockrequests:
            mockrequests.codes = requests.codes
            mockresponse = mockrequests.get.return_value
            mockresponse.status_code = requests.codes.ok
            mockresponse.json.return_value = data
            pres = IIIFPresentation.from_url(manifest_url)
            assert pres.type == "sc:Manifest"
            mockrequests.get.assert_called_with(manifest_url)
            mockrequests.get.return_value.json.assert_called_with()

            # error handling
            # bad status code response on the url
            with pytest.raises(IIIFException) as excinfo:
                mockresponse.status_code = requests.codes.forbidden
                mockresponse.reason = "Forbidden"
                IIIFPresentation.from_url(manifest_url)
            assert "Error retrieving manifest" in str(excinfo.value)
            assert "403 Forbidden" in str(excinfo.value)

            # valid http response but not a json response
            with pytest.raises(IIIFException) as excinfo:
                mockresponse.status_code = requests.codes.ok
                # content type header does not indicate json
                mockresponse.headers = {"content-type": "text/html"}
                mockresponse.json.side_effect = json.decoder.JSONDecodeError(
                    "err", "doc", 1
                )
                IIIFPresentation.from_url(manifest_url)
            assert "No JSON found" in str(excinfo.value)

            # json parsing error
            with pytest.raises(IIIFException) as excinfo:
                # content type header indicates json, but parsing failed
                mockresponse.headers = {"content-type": "application/json"}
                mockresponse.json.side_effect = json.decoder.JSONDecodeError(
                    "err", "doc", 1
                )
                IIIFPresentation.from_url(manifest_url)
            assert "Error parsing JSON" in str(excinfo.value)

    def test_from_url_or_file(self):
        with patch.object(IIIFPresentation, "from_url") as mock_from_url:
            # local fixture file
            pres = IIIFPresentation.from_file_or_url(self.test_manifest)
            assert pres.type == "sc:Manifest"
            mock_from_url.assert_not_called()

            pres = IIIFPresentation.from_file_or_url("http://mani.fe/st")
            mock_from_url.assert_called_with("http://mani.fe/st")

            # nonexistent file path
            with pytest.raises(IIIFException) as excinfo:
                IIIFPresentation.from_file_or_url("/manifest/not/found")
            assert "File not found: " in str(excinfo.value)

    def test_short_id(self):
        manifest_uri = "https://ii.if/resources/p0c484h74c/manifest"
        assert IIIFPresentation.short_id(manifest_uri) == "p0c484h74c"
        canvas_uri = "https://ii.if/resources/p0c484h74c/manifest/canvas/ps7527b878"
        assert IIIFPresentation.short_id(canvas_uri) == "ps7527b878"

    def test_toplevel_attrs(self):
        pres = IIIFPresentation.from_file(self.test_manifest)
        assert pres.context == "http://iiif.io/api/presentation/2/context.json"
        assert (
            pres.id
            == "https://plum.princeton.edu/concern/scanned_resources/ph415q7581/manifest"
        )
        assert pres.type == "sc:Manifest"
        assert pres.label[0] == "Chto my stroim : Tetrad\u02b9 s kartinkami"
        assert pres.viewingHint == "paged"
        assert pres.viewingDirection == "left-to-right"

    def test_nested_attrs(self):
        pres = IIIFPresentation.from_file(self.test_manifest)
        assert isinstance(pres.sequences, list)
        assert (
            pres.sequences[0].id
            == "https://plum.princeton.edu/concern/scanned_resources/ph415q7581/manifest/sequence/normal"
        )
        assert pres.sequences[0].type == "sc:Sequence"
        assert isinstance(pres.sequences[0].canvases, list)
        assert (
            pres.sequences[0].canvases[0].id
            == "https://plum.princeton.edu/concern/scanned_resources/ph415q7581/manifest/canvas/p02871v98d"
        )

    def test_set(self):
        pres = IIIFPresentation.from_file(self.test_manifest)
        pres.label = "New title"
        pres.type = "sc:Collection"
        assert pres.label == "New title"
        assert pres.type == "sc:Collection"

    def test_del(self):
        pres = IIIFPresentation.from_file(self.test_manifest)
        del pres.label
        del pres.type
        assert not hasattr(pres, "label")
        assert not hasattr(pres, "type")

        # accessing missing keys as item vs accessing as attribute
        # currently results in different errors
        # (accurate, but potentially confusing?)

        with pytest.raises(KeyError):
            assert not pres["label"]
        with pytest.raises(KeyError):
            assert not pres["type"]

        with pytest.raises(AttributeError):
            assert not pres.label
        with pytest.raises(AttributeError):
            assert not pres.type

    def test_first_label(self):
        pres = IIIFPresentation.from_file(self.test_manifest)
        assert pres.first_label == pres.label[0]
        pres.label = "unlisted single title"
        assert pres.first_label == pres.label



import json
import os.path
import urllib

import addict

import requests


class IIIFException(Exception):
    """Custom exception for IIIF errors"""


def get_iiif_url(url):
    """Wrapper around :meth:`requests.get` to support conditionally
    adding an auth tokens or other parameters."""
    request_options = {}
    # TODO: need some way of configuring hooks for e.g. setting auth tokens
    return requests.get(url, **request_options)


class IIIFPresentation(addict.Dict):
    """:class:`addict.Dict` subclass for read access to IIIF Presentation
    content"""

    # TODO: document sample use, e.g. @ fields

    at_fields = ["type", "id", "context"]

    @classmethod
    def from_file(cls, path):
        """Initialize :class:`IIIFPresentation` from a file."""
        with open(path) as manifest:
            data = json.loads(manifest.read())
        return cls(data)

    @classmethod
    def from_url(cls, uri):
        """Iniitialize :class:`IIIFPresentation` from a URL.

        :raises: :class:`IIIFException` if URL is not retrieved successfully,
            if the response is not JSON content, or if the JSON cannot be parsed.
        """
        response = get_iiif_url(uri)
        if response.status_code == requests.codes.ok:
            try:
                return cls(response.json())
            except json.decoder.JSONDecodeError as err:
                # if json fails, two possibilities:
                # - we didn't actually get json (e.g. redirect for auth)
                if "application/json" not in response.headers["content-type"]:
                    raise IIIFException("No JSON found at %s" % uri)
                # - there is something wrong with the json
                raise IIIFException("Error parsing JSON for %s: %s" % (uri, err))

        raise IIIFException(
            "Error retrieving manifest at %s: %s %s"
            % (uri, response.status_code, response.reason)
        )

    @classmethod
    def is_url(cls, url):
        """Utility method to check if a path is a url or file"""
        return urllib.parse.urlparse(url).scheme != ""

    @classmethod
    def from_file_or_url(cls, path):
        """Iniitialize :class:`IIIFPresentation` from a file or a url."""
        if os.path.isfile(path):
            return cls.from_file(path)
        elif cls.is_url(path):
            return cls.from_url(path)
        else:
            raise IIIFException("File not found: %s" % path)

    @classmethod
    def short_id(cls, uri):
        """Generate a short id from full manifest/canvas uri identifiers
        for use in local urls.  Logic is based on the recommended
        url pattern from the IIIF Presentation 2.0 specification."""

        # shortening should work reliably for uris that follow
        # recommended url patterns from the spec
        # http://iiif.io/api/presentation/2.0/#a-summary-of-recommended-uri-patterns
        #   manifest:  {scheme}://{host}/{prefix}/{identifier}/manifest
        #   canvas: {scheme}://{host}/{prefix}/{identifier}/canvas/{name}

        # remove trailing /manifest at the end of the url, if present
        if uri.endswith("/manifest"):
            uri = uri[: -len("/manifest")]
        # split on slashes and return the last portion
        return uri.split("/")[-1]

    def __missing__(self, key):
        raise KeyError(self._key(key))

    def _key(self, key):
        # convert key to @key if in the list of fields that requires it
        if key in self.at_fields:
            key = "@%s" % key
        return key

    def __getattr__(self, key):
        try:
            # addict getattr just calls getitem
            return super().__getattr__(self._key(key))
        except KeyError:
            # python hasattr checks for attribute error
            # translate key error to attribute error,
            # since in an attr dict it's kind of both
            raise AttributeError

    def __getitem__(self, key):
        """
        Access a value associated with a key.
        """
        val = super().__getitem__(self._key(key))
        return val

    def __setitem__(self, key, value):
        """
        Add a key-value pair to the instance.
        """
        return super().__setitem__(self._key(key), value)

    def __delitem__(self, key):
        """
        Delete a key-value pair
        """
        super().__delitem__(self._key(key))

    @property
    def first_label(self):
        # label can be a string or list of strings
        if isinstance(self.label, str):
            return self.label
        else:
            return self.label[0]



# iiifclient
# -*- coding: utf-8 -*-

from collections import OrderedDict
from urllib.parse import urlparse

from cached_property import cached_property
import requests


class IIIFImageClientException(Exception):
    """IIIFImageClient custom exception class"""


class ParseError(IIIFImageClientException):
    """Exception raised when an IIIF image could not be parsed"""


# NOTE: possible image component base class?
# commonalities so far: setting defaults on init / parse
# handling exact matches like full/square? (but maybe only region/size),
# validating options (and could add option type checking)


class ImageRegion(object):
    """IIIF Image region. Intended to be used with :class:`IIIFImageClient`.
    Can be initialized with related image object and region options.

    When associated with an image, region is callable and will return
    an updated image object with the modified region options.

    :param img: :class:`IIFImageClient`
    :param full: full region, defaults to true
    :param square: square region, defaults to false
    :param x: x coordinate
    :param y: y coordinate
    :param width: region width
    :param height: region height
    :param percent: region is a percentage
    """

    # region options
    options = OrderedDict(
        [
            ("full", False),
            ("square", False),
            ("x", None),
            ("y", None),
            ("width", None),
            ("height", None),
            ("percent", False),
        ]
    )

    region_defaults = {
        "full": True,
        "square": False,
        "x": None,
        "y": None,
        "width": None,
        "height": None,
        "percent": False,
    }

    coords = ["x", "y", "width", "height"]

    def __init__(self, img=None, **options):
        self.img = img
        self.options = self.region_defaults.copy()
        if options:
            self.set_options(**options)

    def __call__(self, **options):
        if self.img is not None:
            img = self.img.get_copy()
            img.region.set_options(**options)
            return img

    def set_options(self, **options):
        """Update region options.  Same parameters as initialiation."""
        allowed_options = list(self.options.keys())
        # error if an unrecoganized option is specified
        for key in options:
            if key not in allowed_options:
                raise IIIFImageClientException("Unknown option: %s" % key)

        # error if some but not all coordinates are specified
        # or if percentage is specified but not all coordinates are present
        if (
            any([coord in options for coord in self.coords]) or "percent" in options
        ) and not all([coord in options for coord in self.coords]):
            # partial region specified
            raise IIIFImageClientException("Incomplete region specified")

        # TODO: do we need to type checking? bool/int/float?

        self.options.update(**options)
        # if any non-full value is specified, set full to false
        # NOTE: if e.g. square is specified but false, this is wrong
        allowed_options.remove("full")
        if any([(key in allowed_options and options[key]) for key in options.keys()]):
            self.options["full"] = False

    def as_dict(self):
        """Return region options as a dictionary"""
        return self.options

    def __str__(self):
        """Render region information in IIIF region format"""
        if self.options["full"]:
            return "full"
        if self.options["square"]:
            return "square"

        coords = "%(x)g,%(y)g,%(width)g,%(height)g" % self.options
        if self.options["percent"]:
            return "pct:%s" % coords

        return coords

    def parse(self, region):
        """Parse an IIIF Image region string and update the current region"""

        # reset to defaults before parsing
        self.options = self.region_defaults.copy()

        # full?
        if region == "full":
            self.options["full"] = True
            # return immediately
            return

        self.options["full"] = False

        if region == "square":
            self.options["square"] = True
            # return immediately
            return self

        # percent?
        if "pct" in region:
            self.options["percent"] = True
            region = region.split("pct:")[1]

        # split to dictionary
        # if percentage type, cast to float
        try:
            if self.options["percent"]:
                coords = [float(region_c) for region_c in region.split(",")]
            # else, force int
            else:
                coords = [int(region_c) for region_c in region.split(",")]
        except ValueError:
            # failure converting to integer or float
            raise ParseError("Invalid region coordinates: %s" % region)

        if len(coords) != 4:
            raise ParseError("Invalid region coordinates: %s" % region)

        x, y, width, height = coords
        self.options.update({"x": x, "y": y, "width": width, "height": height})

    def canonicalize(self):
        """Canonicalize the current region options so that
        serialization results in canonical format."""
        # From the spec:
        #   “full” if the whole image is requested, (including a “square”
        #    region of a square image), otherwise the x,y,w,h syntax.

        if self.options["full"]:
            # nothing to do
            return

        # if already in x,y,w,h format - nothing to do
        if not any(
            [self.options["full"], self.options["square"], self.options["percent"]]
        ):
            return

        # possbly an error here for every other case if self.img is not set,
        # since it's probably not possible to canonicalize without knowing
        # image size  (any exceptions?)
        if self.img is None:
            raise IIIFImageClientException("Cannot canonicalize without image")

        if self.options["square"]:
            # if image is square, then return full
            if self.img is not None and self.img.image_width == self.img.image_height:
                self.options["full"] = True
                self.options["square"] = False
                return

            # otherwise convert to x,y,w,h
            # from the spec:
            #  The region is defined as an area where the width and height
            #  are both equal to the length of the shorter dimension of the
            #  complete image. The region may be positioned anywhere in the
            #  longer dimension of the image content at the server’s
            #  discretion, and centered is often a reasonable default.

            # determine size of the short edge
            short_edge = min(self.img.image_width, self.img.image_height)
            width = height = short_edge
            # calculate starting long edge point to center the square
            if self.img.image_height == short_edge:
                y = 0
                x = (self.img.image_width - short_edge) / 2
            else:
                x = 0
                y = (self.img.image_height - short_edge) / 2

            self.options.update(
                {"x": x, "y": y, "width": width, "height": height, "square": False}
            )

        if self.options["percent"]:
            # convert percentages to x,y,w,h
            # From the spec:
            #  The region to be returned is specified as a sequence of
            #  percentages of the full image’s dimensions, as reported in
            #  the image information document. Thus, x represents the number
            #  of pixels from the 0 position on the horizontal axis, calculated
            #  as a percentage of the reported width. w represents the width
            #  of the region, also calculated as a percentage of the reported
            #  width. The same applies to y and h respectively. These may be
            #  floating point numbers.

            # convert percentages to dimensions based on image size
            self.options.update(
                {
                    "percent": False,
                    "x": int((self.options["x"] / 100) * self.img.image_width),
                    "y": int((self.options["y"] / 100) * self.img.image_height),
                    "width": int((self.options["width"] / 100) * self.img.image_width),
                    "height": int(
                        (self.options["height"] / 100) * self.img.image_height
                    ),
                }
            )
            return


class ImageSize(object):
    """IIIF Image Size.  Intended to be used with :class:`IIIFImageClient`.
    Can be initialized with related image object and size options.

    When associated with an image, size is callable and will return
    an updated image object with the modified size.

    :param img: :class:`IIFImageClient`
    :param width: optional width
    :param height: optional height
    :param percent: optional percent
    :param exact: size should be exact (boolean, optional)
    """

    # size options
    options = OrderedDict(
        [
            ("full", False),
            ("max", False),
            ("width", None),
            ("height", None),
            ("percent", None),
            ("exact", False),
        ]
    )

    # NOTE: full is being deprecated and replaced with max;
    # full is deprecated in 2.1 and will be removed for 3.0
    # Eventually piffle will need to address that, maybe with some kind of
    # support for selecting a particular version of the IIIF image spec.
    # For now, default size is still full, and max and full are treated as
    # separate modes.  A parsed url with max will return max, and a parsed
    # url with full will return full, but that will probably change
    # once the deprecated full is handled properly.

    size_defaults = {
        "full": True,
        "max": False,
        "width": None,
        "height": None,
        "percent": None,
        "exact": False,
    }

    def __init__(self, img=None, **options):
        self.img = img
        self.options = self.size_defaults.copy()
        if options:
            self.set_options(**options)

    def __call__(self, **options):
        if self.img is not None:
            img = self.img.get_copy()
            img.size.set_options(**options)
            return img

    def set_options(self, **options):
        """Update size options.  Same parameters as initialiation."""
        allowed_options = list(self.options.keys())
        # error if an unrecoganized option is specified
        for key in options:
            if key not in allowed_options:
                raise IIIFImageClientException("Unknown option: %s" % key)

        # TODO: do we need to type checking? bool/int/float?

        self.options.update(**options)
        # if any non-full value is specified, set full to false
        # NOTE: if e.g. square is specified but false, this is wrong
        allowed_options.remove("full")
        if any([key in allowed_options and options[key] for key in options.keys()]):
            self.options["full"] = False

    def as_dict(self):
        """Return size options as a dictionary"""
        return self.options

    def __str__(self):
        if self.options["full"]:
            return "full"
        if self.options["max"]:
            return "max"
        if self.options["percent"]:
            return "pct:%g" % self.options["percent"]

        size = "%s,%s" % (self.options["width"] or "", self.options["height"] or "")
        if self.options["exact"]:
            return "!%s" % size
        return size

    def parse(self, size):
        # reset to defaults before parsing
        self.options = self.size_defaults.copy()

        # full?
        if size == "full":
            self.options["full"] = True
            return

        # for any other case, full should be false
        self.options["full"] = False

        # max?
        if size == "max":
            self.options["max"] = True
            return

        # percent?
        if "pct" in size:
            try:
                self.options["percent"] = float(size.split(":")[1])
                return
            except ValueError:
                raise ParseError("Error parsing size: %s" % size)

        # exact?
        if size.startswith("!"):
            self.options["exact"] = True
            size = size.lstrip("!")

        # split width and height
        width, height = size.split(",")
        try:
            if width != "":
                self.options["width"] = int(width)
            if height != "":
                self.options["height"] = int(height)
        except ValueError:
            raise ParseError("Error parsing size: %s" % size)

    def canonicalize(self):
        """Canonicalize the current size options so that
        serialization results in canonical format."""
        # From the spec:
        #   “full” if the default size is requested,
        #   the w, syntax for images that should be scaled maintaining the
        #   aspect ratio, and the w,h syntax for explicit sizes that change
        #   the aspect ratio.
        #   Note: The size keyword “full” will be replaced with “max” in
        #   version 3.0

        if self.options["full"]:
            # nothing to do
            return

        # possbly an error here for every other case if self.img is not set,
        # since it's probably not possible to canonicalize without knowing
        # image size  (any exceptions?)
        if self.img is None:
            raise IIIFImageClientException("Cannot canonicalize without image")

        if self.options["percent"]:
            # convert percentage to w,h
            scale = self.options["percent"] / 100
            self.options.update(
                {
                    "height": int(self.img.image_height * scale),
                    "width": int(self.img.image_width * scale),
                    "percent": None,
                }
            )
            return

        if self.options["exact"]:
            # from the spec:
            #   The image content is scaled for the best fit such that the
            #   resulting width and height are less than or equal to the
            #   requested width and height. The exact scaling may be determined
            #   by the service provider, based on characteristics including
            #   image quality and system performance. The dimensions of the
            #   returned image content are calculated to maintain the aspect
            #   ratio of the extracted region.

            # determine which edge results in a smaller scale
            wscale = float(self.options["width"]) / float(self.img.image_width)
            hscale = float(self.options["height"]) / float(self.img.image_height)
            scale = min(wscale, hscale)
            # use that scale on original image size, to preserve
            # original aspect ratio
            self.options.update(
                {
                    "exact": False,
                    "height": int(scale * self.img.image_height),
                    "width": int(scale * self.img.image_width),
                }
            )
            return

        # if height only is specified (,h), convert to width only (w,)
        if self.options["height"] and self.options["width"] is None:
            # determine the scale in use, and convert from
            # height only to width only
            scale = float(self.options["height"]) / float(self.img.image_height)
            self.options.update(
                {"height": None, "width": int(scale * self.img.image_width)}
            )


class ImageRotation(object):
    """IIIF Image rotation Intended to be used with :class:`IIIFImageClient`.
    Can be initialized with related image object and rotation options.

    When associated with an image, rotation is callable and will return
    an updated image object with the modified rotatoin options.

    :param img: :class:`IIFImageClient`
    :param degrees: degrees rotation, optional
    :param mirrored: image should be mirrored (boolean, optional, default
       is False)
    """

    # rotation options
    options = OrderedDict(
        [
            ("degrees", None),
            ("mirrored", False),
        ]
    )

    rotation_defaults = {"degrees": 0, "mirrored": False}

    def __init__(self, img=None, **options):
        self.img = img
        self.options = self.rotation_defaults.copy()
        if options:
            self.set_options(**options)

    def __call__(self, **options):
        if self.img is not None:
            img = self.img.get_copy()
            img.rotation.set_options(**options)
            return img

    def set_options(self, **options):
        """Update size options.  Same parameters as initialiation."""
        allowed_options = self.options.keys()
        # error if an unrecoganized option is specified
        for key in options:
            if key not in allowed_options:
                raise IIIFImageClientException("Unknown option: %s" % key)

        # TODO: do we need to type checking? bool/int/float?

        self.options.update(**options)

    def as_dict(self):
        """Return rotation options as a dictionary"""
        return self.options

    def __str__(self):
        return "%s%g" % (
            "!" if self.options["mirrored"] else "",
            self.options["degrees"],
        )

    def parse(self, rotation):
        # reset to defaults before parsing
        self.options = self.rotation_defaults.copy()

        if str(rotation).startswith("!"):
            self.options["mirrored"] = True
            rotation = rotation.lstrip("!")

        # rotation allows float
        self.options["degrees"] = float(rotation)

    def canonicalize(self):
        """Canonicalize the current region options so that
        serialization results in canonical format."""
        # NOTE: explicitly including a canonicalize as a method to make it
        # clear that this field supports canonicalization, but no work
        # is needed since the existing render does the right things:
        # - trim any trailing zeros in a decimal value
        # - leading zero if less than 1
        # - ! if mirrored, followed by integer if possible
        return


class IIIFImageClient(object):
    """Simple IIIF Image API client for generating IIIF image urls
     in an object-oriented, pythonic fashion.  Can be extended,
     when custom logic is needed to set the image id.  Provides
     a fluid interface, so that IIIF methods can be chained, e.g.::

         iiif_img.size(width=300).rotation(90).format('png')

    Note that this returns a new image instance with the specified
    options, and the original image will remain unchanged.

     .. Note::

         Method to set quality not yet available.
    """

    api_endpoint = None
    image_id = None
    default_format = "jpg"

    # iiif defaults for each sections
    image_defaults = {
        "quality": "default",  # color, gray, bitonal, default
        "fmt": default_format,
    }
    allowed_formats = ["jpg", "tif", "png", "gif", "jp2", "pdf", "webp"]

    def __init__(
        self,
        api_endpoint=None,
        image_id=None,
        region=None,
        size=None,
        rotation=None,
        quality=None,
        fmt=None,
    ):
        self.image_options = self.image_defaults.copy()
        # NOTE: using underscore to differenteate objects from methods
        # but it could be reasonable to make objects public
        self.region = ImageRegion(self)
        self.size = ImageSize(self)
        self.rotation = ImageRotation(self)

        if api_endpoint is not None:
            # remove any trailing slash to avoid duplicate slashes
            self.api_endpoint = api_endpoint.rstrip("/")

        # FIXME: image_id is not required on init to allow subclassing
        # and customizing via get_image_id, but should probably cause
        # an error if you attempt to serialize the url and it is not set
        # (same for a few other options, probably, too...)
        if image_id is not None:
            self.image_id = image_id

        # for now, if region option is specified parse as string
        if region is not None:
            self.region.parse(region)
        if size is not None:
            self.size.parse(size)
        if rotation is not None:
            self.rotation.parse(rotation)

        if quality is not None:
            self.image_options["quality"] = quality
        if fmt is not None:
            self.image_options["fmt"] = fmt

    def get_image_id(self):
        "Image id to be used in contructing urls"
        return self.image_id

    def __str__(self):
        info = self.image_options.copy()
        info.update(
            {
                "endpoint": self.api_endpoint,
                "id": self.get_image_id(),
                "region": str(self.region),
                "size": str(self.size),
                "rot": str(self.rotation),
            }
        )
        return (
            "%(endpoint)s/%(id)s/%(region)s/%(size)s/%(rot)s/%(quality)s.%(fmt)s" % info
        )

    def __repr__(self):
        return "<IIIFImageClient %s>" % self.get_image_id()
        # include non-defaults?

    def info(self):
        "JSON info url"
        return "%(endpoint)s/%(id)s/info.json" % {
            "endpoint": self.api_endpoint,
            "id": self.get_image_id(),
        }

    @cached_property
    def image_info(self):
        "Retrieve image information provided as JSON at info url"
        resp = requests.get(self.info())
        if resp.status_code == requests.codes.ok:
            return resp.json()

        resp.raise_for_status()

    @property
    def image_width(self):
        "Image width as reported in :attr:`image_info`"
        return self.image_info["width"]

    @property
    def image_height(self):
        "Image height as reported in :attr:`image_info`"
        return self.image_info["height"]

    def get_copy(self):
        "Get a clone of the current settings for modification."
        clone = self.__class__(self.api_endpoint, self.image_id, **self.image_options)
        # copy region, size, and rotation - no longer included in
        # image_options dict
        clone.region.set_options(**self.region.as_dict())
        clone.size.set_options(**self.size.as_dict())
        clone.rotation.set_options(**self.rotation.as_dict())
        return clone

    # method to set quality not yet implemented

    def format(self, image_format):
        "Set output image format"
        if image_format not in self.allowed_formats:
            raise IIIFImageClientException("Image format %s unknown" % image_format)
        img = self.get_copy()
        img.image_options["fmt"] = image_format
        return img

    def canonicalize(self):
        """Canonicalize the URI"""
        img = self.get_copy()
        img.region.canonicalize()
        img.size.canonicalize()
        img.rotation.canonicalize()
        return img

    @classmethod
    def init_from_url(cls, url):
        """Init ImageClient using Image API parameters from URI.  Detect
        image vs. info request. Can count reliably from the end of the URI
        backwards, but cannot assume how many slashes make up the api_endpoint.
        Returns new instance of IIIFImageClient.
        Per http://iiif.io/api/image/2.0/#image-request-uri-syntax, using
        slashes to parse URI"""

        # first parse as a url
        parsed_url = urlparse(url)
        # then split the path on slashes
        # and remove any empty strings
        path_components = [path for path in parsed_url.path.split("/") if path]
        if not path_components:
            raise ParseError("Invalid IIIF image url: %s" % url)
        # pop off last portion of the url to determine if this is an info url
        path_basename = path_components.pop()
        opts = {}

        # info request
        if path_basename == "info.json":
            # NOTE: this is unlikely to happen; more likely, if information is
            # missing, we will misinterpret the api endpoint or the image id
            if len(path_components) < 1:
                raise ParseError("Invalid IIIF image information url: %s" % url)
            image_id = path_components.pop()

        # image request
        else:
            # check for enough IIIF parameters
            if len(path_components) < 4:
                raise ParseError("Invalid IIIF image request: %s" % url)

            # pop off url portions as they are used so we can easily
            # make use of leftover path to reconstruct the api endpoint
            quality, fmt = path_basename.split(".")
            rotation = path_components.pop()
            size = path_components.pop()
            region = path_components.pop()
            image_id = path_components.pop()
            opts.update(
                {
                    "region": region,
                    "size": size,
                    "rotation": rotation,
                    "quality": quality,
                    "fmt": fmt,
                }
            )

        # construct the api endpoint url from the parsed url and whatever
        # portions of the url path are leftover
        # remove empty strings from the remaining path components
        path_components = [p for p in path_components if p]
        api_endpoint = "%s://%s/%s" % (
            parsed_url.scheme,
            parsed_url.netloc,
            "/".join(path_components) if path_components else "",
        )

        # init and return instance
        return cls(api_endpoint=api_endpoint, image_id=image_id, **opts)

    def as_dict(self):
        """
        Dictionary of with all image request options.
        request parameters. Returns a dictionary with all image request
        parameters parsed to their most granular level. Can be helpful
        for acting logically on particular request parameters like height,
        width, mirroring, etc.
        """
        return OrderedDict(
            [
                ("region", self.region.as_dict()),
                ("size", self.size.as_dict()),
                ("rotation", self.rotation.as_dict()),
                ("quality", self.image_options["quality"]),
                ("format", self.image_options["fmt"]),
            ]
        )
